/*  lib.c --- igorpluc access library

 Copyright (C) 2009 by Walter C. Pelissero

 Author: Walter C. Pelissero <walter@pelissero.de>
 Project: igorplug

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2, or (at
your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.*/

#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/ioctl.h>
#include <dev/usb/usb.h>
#include <string.h>

#include "igorplug.h"

#define READ_ENDPOINT 1
#define WRITE_ENDPOINT 2
#define SET_INFRABUFFER_EMPTY 1
#define GET_INFRACODE 2
#define DEVICE_HEADERLEN        2
/* it appears the IgorPlug buffer is limited to 36 bytes */
#define DEVICE_BUFLEN           (36 + DEVICE_HEADERLEN)
#define USB_CTRL_GET_TIMEOUT    5000

/* This is also in util.h, but we shouldn't use that include. */
typedef enum {false, true} bool;


static int
usb_set_timeout (int fd, int timeout)
{
  return ioctl(fd, USB_SET_TIMEOUT, &timeout) >= 0;
}

static int
usb_do_request (int fd, struct usb_ctl_request *req)
{
  return ioctl(fd, USB_DO_REQUEST, req) >= 0;
}

static int
usb_control_msg(int fd, int requesttype, int request, int value,
		int index, char *bytes, int size, int timeout)
{
  struct usb_ctl_request req;

  req.ucr_request.bmRequestType = requesttype;
  req.ucr_request.bRequest = request;
  USETW(req.ucr_request.wValue, value);
  USETW(req.ucr_request.wIndex, index);
  USETW(req.ucr_request.wLength, size);

  req.ucr_data = bytes;
  req.ucr_flags = USBD_SHORT_XFER_OK;

  if (usb_set_timeout(fd, timeout))
    {
      if (usb_do_request(fd, &req))
	return UGETW(req.ucr_request.wLength);
    }
  return -1;
}

static int
clear_buffer (int fd)
{
  char buf[1];
  
  return usb_control_msg(fd,
			 UT_VENDOR | UE_DIR_IN,
			 SET_INFRABUFFER_EMPTY,
			 0, 0,
			 buf, sizeof(buf),
			 USB_CTRL_GET_TIMEOUT);
}

static int
read_buffer (int fd, void *buf, unsigned length)
{
  return usb_control_msg(fd, UT_VENDOR | UE_DIR_IN,
			 GET_INFRACODE,
			 0, 0, 
			 (char *) buf, length, 
			 USB_CTRL_GET_TIMEOUT);
}

/* Receive the content of IgorPlug's buffer.  Keep looping until the
   buffer is ready with some meaningful data. It doesn't guarantee
   that the receiver didn't pick some noise, it just guarantees that
   the packet is suitable for further processing. */
int
igorplug_read_buffer (int fd, IRbuffer buf)
{
  if (clear_buffer(fd) < 0)
    return -1;
  do
    {
      if (read_buffer(fd, buf, DEVICE_BUFLEN) < 0)
	return -1;
    }
  while (IGORPLUG_DATA_LENGTH(buf) == 0 ||
	 IGORPLUG_DATA_LENGTH(buf) > (DEVICE_BUFLEN - DEVICE_HEADERLEN));
  return 0;
}

bool
compare_IR_stretch (IRbyte x, IRbyte y, unsigned tolerance)
{
  if (abs((int) x - (int) y) <= tolerance)
    return true;
  return false;
}

bool
compare_codes (IRbyte *c1, IRbyte *c2, unsigned length, unsigned tolerance)
{
  int i;

  for (i = 0; i < length; ++i)
    if (!compare_IR_stretch(c1[i], c2[i], tolerance))
      return false;
  return true;
}

bool
compare_buffers (const IRbuffer buf1, const IRbuffer buf2, unsigned tolerance)
{
  if (IGORPLUG_DATA_LENGTH(buf1) != IGORPLUG_DATA_LENGTH(buf2) ||
      abs((int) IGORPLUG_SERIAL(buf1) - (int) IGORPLUG_SERIAL(buf2)) != 1)
    return 1;
  if (!compare_codes(buf1 + DEVICE_HEADERLEN, buf2 + DEVICE_HEADERLEN, IGORPLUG_DATA_LENGTH(buf1), tolerance))
    return false;
  /* they match */
  return true;
}

/* Like igorplug_read_buffer, but returns only when two (almost)
   identical packets have been seen.  The tolerance argument tells how
   similar two packets must be to be accepted.  The second packet is
   returned in buf.  */
int
igorplug_read_buffer2 (int fd, IRbuffer buf, unsigned tolerance)
{
  unsigned char buf2[DEVICE_BUFLEN];

  IGORPLUG_DATA_LENGTH(buf) = 0;
  do
    {
      memcpy(buf2, buf, sizeof(buf2));
      if (igorplug_read_buffer(fd, buf) < 0)
	return -1;
    }
  while (!compare_buffers(buf2, buf, tolerance));
  return 0;
}

/* Check that it's actualy an IgorPlug and not your hair drier. */
static int
check_device (int fd)
{
  struct usb_device_info di;

  if (ioctl(fd, USB_GET_DEVICEINFO, &di) < 0)
    return 0;
  if (di.udi_productNo == 2 && di.udi_vendorNo == 0x3EB)
    return 1;
  return 0;
}

IRbuffer
igorplug_alloc_buffer ()
{
  return malloc(DEVICE_BUFLEN);
}

/* Open the supplied ugen device and make sure it's an IgorPlug */
int
igorplug_open_device (const char *device_pathname)
{
  int fd;

  if ((fd = open (device_pathname, O_RDWR)) >= 0)
    {
      if (check_device(fd))
	return fd;
    }
  close(fd);
  return -1;
}
